!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   at_rpath_fep_mod
!> @brief   Free-energy perturbation along given MEP
!! @authors Yoshinobu Akinaga (YA), Kiyoshi Yagi (KY)
! 
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module at_rpath_fep_mod

  use at_energy_mod
  use at_md_leapfrog_mod
  use at_md_vverlet_mod
  use at_dynamics_str_mod
  use at_dynvars_str_mod
  use at_dynvars_mod
  use at_ensemble_str_mod
  use at_rpath_str_mod
  use at_rpath_mep_mod
  use at_constraints_str_mod
  use at_boundary_str_mod
  use at_boundary_mod
  use at_pairlist_str_mod
  use at_pairlist_mod
  use at_enefunc_str_mod
  use at_output_str_mod
  use at_output_mod
  use fileio_rst_mod
  use fileio_rstmep_mod
  use molecules_str_mod
  use select_mod
  use messages_mod
  use mpi_parallel_mod
  use constants_mod
#ifdef MPI
  use mpi
#endif

  implicit none
  private

  ! subroutines
  public  :: setup_rpath_fep
  public  :: run_rpath_fep
  private :: calc_fep

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_rpath_fep
  !> @brief        setup RPATH for FEP
  !! @authors      YA, KY
  !! @param[in]    fep_period : the period of FEP calculation
  !! @param[in]    mepatm_select_index : index of MEP atoms
  !! @param[in]    sel_info   : selector input information
  !! @param[in]    rst        : restart information
  !! @param[in]    rstmep     : rstmep data
  !! @param[in]    molecule   : molecule information
  !! @param[in]    qmmm       : qmmm     information
  !! @param[in]    dynamics   : dynamics information
  !! @param[inout] dynvars    : dynamic variables
  !! @param[inout] rpath      : RPATH information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_rpath_fep(fep_period, mepatm_select_index, sel_info, rst, &
                             rstmep, molecule, qmmm, dynamics, dynvars, rpath)

    ! formal arguments
    integer,                 intent(in)    :: fep_period
    character(256),          intent(in)    :: mepatm_select_index
    type(s_sel_info),        intent(in)    :: sel_info
    type(s_rst),     target, intent(in)    :: rst
    type(s_rstmep),          intent(in)    :: rstmep
    type(s_dynamics),target, intent(in)    :: dynamics
    type(s_molecule),        intent(in)    :: molecule
    type(s_qmmm),            intent(inout) :: qmmm
    type(s_dynvars),         intent(inout) :: dynvars
    type(s_rpath),   target, intent(inout) :: rpath

    ! local variables
    integer                    :: ns, i, j, k, m, ii, iatom
    integer                    :: shift, next, id, atomid
    integer                    :: replicaid, errorcode, my_rank, ierr

    return

  end subroutine setup_rpath_fep

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    run_rpath_fep
  !> @brief        run rpath FEP
  !! @authors      YA, KY
  !! @param[inout] output      : output information
  !! @param[inout] molecule    : molecule information
  !! @param[inout] enefunc     : potential energy functions information
  !! @param[inout] dynvars     : dynamics variables information
  !! @param[inout] dynamics    : dynamics information
  !! @param[inout] pairlist    : pairlist information
  !! @param[inout] boundary    : boundary conditions information
  !! @param[inout] constraints : constraints information
  !! @param[inout] ensemble    : ensemble information
  !! @param[inout] rpath       : RPATH information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine run_rpath_fep(output, molecule, enefunc, dynvars, dynamics,   &
                           pairlist, boundary, constraints, ensemble, rpath)

    ! formal arguments
    type(s_output),           intent(inout) :: output
    type(s_molecule), target, intent(inout) :: molecule
    type(s_enefunc),          intent(inout) :: enefunc
    type(s_dynvars),  target, intent(inout) :: dynvars
    type(s_dynamics),         intent(inout) :: dynamics
    type(s_pairlist),         intent(inout) :: pairlist
    type(s_boundary),         intent(inout) :: boundary
    type(s_constraints),      intent(inout) :: constraints
    type(s_ensemble),         intent(inout) :: ensemble
    type(s_rpath),            intent(inout) :: rpath

    ! local variables
    integer                   :: i, j, k
    integer                   :: n, niter
    integer                   :: iloop_start, iloop_end
    integer                   :: replicaid, errorcode, ierr
    integer                   :: ii, iatom
    real(wp)                  :: dene_qm
    real(wp),     allocatable :: work(:,:)


    return

  end subroutine run_rpath_fep

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    calc_fep
  !> @brief        calculate the free-energy using FEP
  !! @authors      YA, KY
  !! @param[inout] molecule : molecule information
  !! @param[inout] enefunc  : potential energy functions information
  !! @param[inout] dynvars  : dynamical variables information
  !! @param[inout] rpath    : RPATH information
  !! @param[inout] pairlist : non-bond pair list information
  !! @param[inout] boundary : boundary conditions information
  !! @param[inout] ensemble : ensemble information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine calc_fep(molecule, enefunc, dynvars, rpath, pairlist, &
                      boundary, ensemble)

    ! formal arguments
    type(s_molecule),         intent(inout) :: molecule
    type(s_enefunc),  target, intent(inout) :: enefunc
    type(s_dynvars),  target, intent(inout) :: dynvars
    type(s_rpath),            intent(inout) :: rpath
    type(s_pairlist),         intent(inout) :: pairlist
    type(s_boundary),         intent(inout) :: boundary
    type(s_ensemble),         intent(inout) :: ensemble

    ! local variables
    type(s_qmmm), pointer :: qmmm
    real(wp),     pointer :: coord(:,:)
    real(wp), allocatable :: mep_coord_tmp(:,:)

    integer    :: replicaid
    integer    :: i, ii, iatom
    real(wp)   :: beta, dene, e1f, e1b, e0
    character(1) :: fcb
    
  end subroutine calc_fep

end module at_rpath_fep_mod

